@@ -1,6 +1,3 @@
-#define _POSIX_C_SOURCE 200809L
-#include <assert.h>
-#include <errno.h>
 #include <getopt.h>
 #include <libgen.h>
 #include <stdlib.h>
@@ -11,6 +8,7 @@
 
 #include "config.h"
 #include "criteria.h"
+#include "string-util.h"
 #include "types.h"
 
 static int32_t max(int32_t a, int32_t b) {
@@ -88,8 +86,12 @@ void init_default_style(struct mako_style *style) {
 	style->padding.bottom = 5;
 	style->padding.left = 5;
 
+	style->border_radius.top = 0;
+	style->border_radius.right = 0;
+	style->border_radius.bottom = 0;
+	style->border_radius.left = 0;
+
 	style->border_size = 2;
-	style->border_radius = 0;
 
 #ifdef HAVE_ICONS
 	style->icons = true;
@@ -98,6 +100,7 @@ void init_default_style(struct mako_style *style) {
 #endif
 	style->max_icon_size = 64;
 	style->icon_path = strdup("");  // hicolor and pixmaps are implicit.
+	style->icon_border_radius = 0;
 
 	style->font = strdup("monospace 10");
 	style->markup = true;
@@ -266,6 +269,11 @@ bool apply_style(struct mako_style *target, const struct mako_style *style) {
 		target->spec.icon_path = true;
 	}
 
+	if (style->spec.icon_border_radius) {
+		target->icon_border_radius = style->icon_border_radius;
+		target->spec.icon_border_radius = true;
+	}
+
 	if (style->spec.font) {
 		free(target->font);
 		target->font = new_font;
@@ -343,7 +351,7 @@ bool apply_style(struct mako_style *target, const struct mako_style *style) {
 		target->spec.icon_location = true;
 	}
 
-	if (style->border_radius) {
+	if (style->spec.border_radius) {
 		target->border_radius = style->border_radius;
 		target->spec.border_radius = true;
 	}
@@ -409,6 +417,7 @@ bool apply_superset_style(
 	target->spec.margin = true;
 	target->spec.padding = true;
 	target->spec.border_size = true;
+	target->spec.border_radius = true;
 	target->spec.icons = true;
 	target->spec.max_icon_size = true;
 	target->spec.default_timeout = true;
@@ -449,6 +458,11 @@ bool apply_superset_style(
 		target->padding.bottom =
 			max(style->padding.bottom, target->padding.bottom);
 		target->padding.left = max(style->padding.left, target->padding.left);
+		target->border_radius.top = max(style->border_radius.top, target->border_radius.top);
+		target->border_radius.right = max(style->border_radius.right, target->border_radius.right);
+		target->border_radius.bottom =
+			max(style->border_radius.bottom, target->border_radius.bottom);
+		target->border_radius.left = max(style->border_radius.left, target->border_radius.left);
 		target->border_size = max(style->border_size, target->border_size);
 		target->icons = style->icons || target->icons;
 		target->max_icon_size = max(style->max_icon_size, target->max_icon_size);
@@ -488,6 +502,25 @@ bool apply_superset_style(
 	return true;
 }
 
+static char *expand_config_path(const char *path) {
+	if (strncmp(path, "/", 1) == 0) {
+		return strdup(path);
+	}
+
+	if (strncmp(path, "~/", 2) != 0) {
+		fprintf(stderr, "Config path must start with / or ~/\n");
+		return NULL;
+	}
+
+	const char *home = getenv("HOME");
+	if (home == NULL) {
+		fprintf(stderr, "HOME env var not set\n");
+		return NULL;
+	}
+
+	return mako_asprintf("%s/%s", home, path + 2);
+}
+
 static bool apply_config_option(struct mako_config *config, const char *name,
 		const char *value) {
 	if (strcmp(name, "sort") == 0) {
@@ -509,6 +542,9 @@ static bool apply_config_option(struct mako_config *config, const char *name,
 		return true;
 	} else if (strcmp(name, "max-history") == 0) {
 		return parse_int(value, &config->max_history);
+	} else if (strcmp(name, "include") == 0) {
+		char *path = expand_config_path(value);
+		return path && load_config_file(config, path) == 0;
 	}
 
 	return false;
@@ -541,8 +577,8 @@ static bool apply_style_option(struct mako_style *style, const char *name,
 	} else if (strcmp(name, "padding") == 0) {
 		spec->padding = parse_directional(value, &style->padding);
 		if (spec->border_radius && spec->padding) {
-			style->padding.left = max(style->border_radius, style->padding.left);
-			style->padding.right = max(style->border_radius, style->padding.right);
+			style->padding.left = max(style->border_radius.left, style->padding.left);
+			style->padding.right = max(style->border_radius.right, style->padding.right);
 		}
 		return spec->padding;
 	} else if (strcmp(name, "border-size") == 0) {
@@ -578,6 +614,9 @@ static bool apply_style_option(struct mako_style *style, const char *name,
 	} else if (strcmp(name, "icon-path") == 0) {
 		free(style->icon_path);
 		return spec->icon_path = !!(style->icon_path = strdup(value));
+	} else if (strcmp(name, "icon-border-radius") == 0) {
+		spec->icon_border_radius = parse_int_ge(value, &style->icon_border_radius, 0);
+		return spec->icon_border_radius;
 	} else if (strcmp(name, "markup") == 0) {
 		return spec->markup = parse_boolean(value, &style->markup);
 	} else if (strcmp(name, "actions") == 0) {
@@ -611,10 +650,10 @@ static bool apply_style_option(struct mako_style *style, const char *name,
 	} else if (strcmp(name, "history") == 0) {
 		return spec->history = parse_boolean(value, &style->history);
 	} else if (strcmp(name, "border-radius") == 0) {
-		spec->border_radius = parse_int_ge(value, &style->border_radius, 0);
+		spec->border_radius = parse_directional(value, &style->border_radius);
 		if (spec->border_radius && spec->padding) {
-			style->padding.left = max(style->border_radius, style->padding.left);
-			style->padding.right = max(style->border_radius, style->padding.right);
+			style->padding.left = max(style->border_radius.left, style->padding.left);
+			style->padding.right = max(style->border_radius.right, style->padding.right);
 		}
 		return spec->border_radius;
 	} else if (strcmp(name, "max-visible") == 0) {
@@ -646,6 +685,8 @@ static bool apply_style_option(struct mako_style *style, const char *name,
 			binding.action = MAKO_BINDING_NONE;
 		} else if (strcmp(value, "dismiss") == 0) {
 			binding.action = MAKO_BINDING_DISMISS;
+		} else if (strcmp(value, "dismiss --no-history") == 0) {
+			binding.action = MAKO_BINDING_DISMISS_NO_HISTORY;
 		} else if (strcmp(value, "dismiss-all") == 0) {
 			binding.action = MAKO_BINDING_DISMISS_ALL;
 		} else if (strcmp(value, "dismiss-group") == 0) {
@@ -699,49 +740,44 @@ static bool file_exists(const char *path) {
 	return path && access(path, R_OK) != -1;
 }
 
-static char *get_config_path(void) {
-	static const char *config_paths[] = {
-		"$HOME/.mako/config",
-		"$XDG_CONFIG_HOME/mako/config",
+static char *get_default_config_path() {
+	const char *home = getenv("HOME");
+	if (home == NULL) {
+		fprintf(stderr, "HOME env var not set\n");
+		return NULL;
+	}
+
+	const char *config_home = getenv("XDG_CONFIG_HOME");
+	char *config_home_fallback = NULL;
+	if (config_home == NULL || config_home[0] == '\0') {
+		config_home_fallback = mako_asprintf("%s/.config", home);
+		config_home = config_home_fallback;
+	}
+
+	char *config_paths[] = {
+		mako_asprintf("%s/.mako/config", home),
+		mako_asprintf("%s/mako/config", config_home),
 	};
 
-	if (!getenv("XDG_CONFIG_HOME")) {
-		char *home = getenv("HOME");
-		if (!home) {
-			return NULL;
-		}
-		char config_home[strlen(home) + strlen("/.config") + 1];
-		strcpy(config_home, home);
-		strcat(config_home, "/.config");
-		setenv("XDG_CONFIG_HOME", config_home, 1);
-	}
-
-	for (size_t i = 0; i < sizeof(config_paths) / sizeof(char *); ++i) {
-		wordexp_t p;
-		if (wordexp(config_paths[i], &p, 0) == 0) {
-			char *path = strdup(p.we_wordv[0]);
-			wordfree(&p);
-			if (file_exists(path)) {
-				return path;
-			}
-			free(path);
+	size_t config_paths_len = sizeof(config_paths) / sizeof(config_paths[0]);
+	char *found_path = NULL;
+	for (size_t i = 0; i < config_paths_len; ++i) {
+		char *path = config_paths[i];
+		if (file_exists(path)) {
+			found_path = strdup(path);
+			break;
 		}
 	}
 
-	return NULL;
+	for (size_t i = 0; i < config_paths_len; ++i) {
+		free(config_paths[i]);
+	}
+	free(config_home_fallback);
+
+	return found_path;
 }
 
-int load_config_file(struct mako_config *config, char *config_arg) {
-	char *path = NULL;
-	if (config_arg == NULL) {
-		path = get_config_path();
-		if (!path) {
-			return 0;
-		}
-	} else {
-		path = config_arg;
-	}
-
+int load_config_file(struct mako_config *config, char *path) {
 	FILE *f = fopen(path, "r");
 	if (!f) {
 		fprintf(stderr, "Unable to open %s for reading\n", path);
@@ -761,17 +797,38 @@ int load_config_file(struct mako_config *config, char *config_arg) {
 		wl_container_of(config->criteria.next, criteria, link);
 
 	size_t n = 0;
-	while (getline(&line, &n, f) > 0) {
+	ssize_t len = 0;
+	while ((len = getline(&line, &n, f)) != -1) {
 		++lineno;
-		if (line[0] == '\0' || line[0] == '\n' || line[0] == '#') {
+
+		// We can't operate on `line` because we need to free it, so
+		// make a copy
+		char *stripped_line = line;
+
+		// Strip leading whitespace
+		while (stripped_line[0] == ' ' || stripped_line[0] == '\t') {
+			++stripped_line;
+			--len;
+		}
+
+		if (stripped_line[0] == '\0' || stripped_line[0] == '\n' ||
+				stripped_line[0] == '#') {
 			continue;
 		}
 
-		if (line[strlen(line) - 1] == '\n') {
-			line[strlen(line) - 1] = '\0';
+		if (stripped_line[len - 1] == '\n') {
+			stripped_line[--len] = '\0';
 		}
 
-		if (line[0] == '[' && line[strlen(line) - 1] == ']') {
+		// Strip trailing whitespace
+		while (stripped_line[len - 1] == ' ' ||
+				stripped_line[len - 1] == '\t') {
+			--len;
+		}
+		stripped_line[len] = '\0';
+
+
+		if (stripped_line[0] == '[' && stripped_line[len - 1] == ']') {
 			// Since we hit the end of the previous criteria section, validate
 			// that it doesn't break any rules before moving on.
 			if (criteria != NULL && !validate_criteria(criteria)) {
@@ -782,7 +839,7 @@ int load_config_file(struct mako_config *config, char *config_arg) {
 			}
 
 			free(section);
-			section = strndup(line + 1, strlen(line) - 2);
+			section = strndup(stripped_line + 1, len - 2);
 			criteria = create_criteria(config);
 			if (!parse_criteria(section, criteria)) {
 				fprintf(stderr, "[%s:%d] Invalid criteria definition\n", base,
@@ -793,7 +850,7 @@ int load_config_file(struct mako_config *config, char *config_arg) {
 			continue;
 		}
 
-		char *eq = strchr(line, '=');
+		char *eq = strchr(stripped_line, '=');
 		if (!eq) {
 			fprintf(stderr, "[%s:%d] Expected key=value\n", base, lineno);
 			ret = -1;
@@ -803,16 +860,18 @@ int load_config_file(struct mako_config *config, char *config_arg) {
 		bool valid_option = false;
 		eq[0] = '\0';
 
-		valid_option = apply_style_option(&criteria->style, line, eq + 1);
+		valid_option = apply_style_option(&criteria->style,
+				stripped_line, eq + 1);
 
 		if (!valid_option && section == NULL) {
-			valid_option = apply_config_option(config, line, eq + 1);
+			valid_option = apply_config_option(config,
+					stripped_line, eq + 1);
 		}
 
-
 		if (!valid_option) {
+			eq[0] = '=';
 			fprintf(stderr, "[%s:%d] Failed to parse option '%s'\n",
-				base, lineno, line);
+				base, lineno, stripped_line);
 			ret = -1;
 			break;
 		}
@@ -853,6 +912,7 @@ int parse_config_arguments(struct mako_config *config, int argc, char **argv) {
 		{"icon-location", required_argument, 0, 0},
 		{"icon-path", required_argument, 0, 0},
 		{"max-icon-size", required_argument, 0, 0},
+		{"icon-border-radius", required_argument, 0, 0},
 		{"markup", required_argument, 0, 0},
 		{"actions", required_argument, 0, 0},
 		{"format", required_argument, 0, 0},
@@ -898,9 +958,12 @@ int parse_config_arguments(struct mako_config *config, int argc, char **argv) {
 		return opt_status;
 	}
 
-	int config_status = load_config_file(config, config_arg);
-	if (config_status < 0) {
-		return -1;
+	char *config_path = config_arg ? config_arg : get_default_config_path();
+	if (config_path) {
+		int config_status = load_config_file(config, config_path);
+		if (config_status < 0) {
+			return -1;
+		}
 	}
 
 	optind = 1;
